""" A plugin event. """


# Enthought library imports.
from traits.api import Instance, Vetoable


class PluginEvent(Vetoable):
    """ A plugin event. """

    # The plugin that the event is for.
    plugin = Instance('envisage.api.IPlugin')

#### EOF ######################################################################



""" Tests for class load hooks. """


from envisage.api import ClassLoadHook
from traits.api import HasTraits
from traits.testing.unittest_tools import unittest


# This module's package.
PKG = 'envisage.tests'


class ClassLoadHookTestCase(unittest.TestCase):
    """ Tests for class load hooks. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_connect(self):
        """ connect """

        def on_class_loaded(cls):
            """ Called when a class is loaded. """

            on_class_loaded.cls = cls

            return

        # To register with 'MetaHasTraits' we use 'module_name.class_name'.
        hook = ClassLoadHook(
            class_name = ClassLoadHookTestCase.__module__ + '.Foo',
            on_load    = on_class_loaded
        )
        hook.connect()

        class Foo(HasTraits):
            pass

        self.assertEqual(Foo, on_class_loaded.cls)

        return

    def test_class_already_loaded(self):
        """ class already loaded """

        def on_class_loaded(cls):
            """ Called when a class is loaded. """

            on_class_loaded.cls = cls

            return

        # To register with 'MetaHasTraits' we use 'module_name.class_name'.
        hook = ClassLoadHook(
            class_name = self._get_full_class_name(ClassLoadHookTestCase),
            on_load    = on_class_loaded
        )
        hook.connect()

        # Make sure the 'on_load' got called immediately because the class is
        # already loaded.
        self.assertEqual(ClassLoadHookTestCase, on_class_loaded.cls)

        return

    def test_disconnect(self):
        """ disconnect """

        def on_class_loaded(cls):
            """ Called when a class is loaded. """

            on_class_loaded.cls = cls

            return

        # To register with 'MetaHasTraits' we use 'module_name.class_name'.
        hook = ClassLoadHook(
            class_name = ClassLoadHookTestCase.__module__ + '.Foo',
            on_load    = on_class_loaded
        )
        hook.connect()

        class Foo(HasTraits):
            pass

        self.assertEqual(Foo, on_class_loaded.cls)

        # 'Reset' the listener,
        on_class_loaded.cls = None

        # Now disconnect.
        hook.disconnect()

        class Foo(HasTraits):
            pass

        self.assertEqual(None, on_class_loaded.cls)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_full_class_name(self, cls):
        """ Return the full (possibly) dotted name of a class. """

        return cls.__module__ + '.' + cls.__name__


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" An  IPython kernel plugin. """

# Enthought library imports.
from envisage.plugins.ipython_kernel.ipython_kernel_plugin import (
    IPYTHON_KERNEL_PROTOCOL)
from envisage.ui.tasks.api import TaskExtension
from envisage.api import Plugin
from traits.api import List
from pyface.tasks.action.api import SGroup, SchemaAddition


TASK_EXTENSIONS = 'envisage.ui.tasks.task_extensions'


class IPythonKernelUIPlugin(Plugin):
    """ Contributes UI actions on top of the IPython Kernel Plugin. """

    #### 'IPlugin' interface ##################################################

    # The plugin unique identifier.
    id = 'envisage.plugins.ipython_kernel_ui'

    # The plugin name (suitable for displaying to the user).
    name = 'IPython embedded kernel UI plugin'

    #### Contributions to extension points made by this plugin ################

    contributed_task_extensions = List(contributes_to=TASK_EXTENSIONS)

    #### Trait initializers ###################################################

    def _contributed_task_extensions_default(self):

        from .actions import StartQtConsoleAction

        def menu_factory():
            kernel = self.application.get_service(IPYTHON_KERNEL_PROTOCOL)
            return SGroup(
                StartQtConsoleAction(kernel=kernel),
                id='ipython'
            )

        return [
            TaskExtension(
                actions=[
                    SchemaAddition(
                        path='MenuBar/View',
                        factory=menu_factory,
                        id='IPythonSchema',
                    ),
                ]
            )
        ]



""" An action set is a collection of menus, groups, and actions. """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import Bool, Dict, HasTraits, List, Str, provides
from traits.util.camel_case import camel_case_to_words

# Local imports.
from .action import Action
from .group import Group
from .menu import Menu
from .tool_bar import ToolBar
from .i_action_set import IActionSet


# Logging.
logger = logging.getLogger(__name__)


@provides(IActionSet)
class ActionSet(HasTraits):
    """ An action set is a collection of menus, groups, and actions. """

    # The action set's globally unique identifier.
    id = Str

    # The action set's name.
    #
    # fixme: This is not currently used, but in future it will be the name that
    # is shown to the user when they are customizing perspectives by adding or
    # removing action sets etc.
    name = Str

    # The actions in this set.
    actions = List(Action)

    # The groups in this set.
    groups = List(Group)

    # The menus in this set.
    menus = List(Menu)

    # The tool bars in this set.
    tool_bars = List(ToolBar)

    # Are the actions and menus in this set enabled (if they are disabled they
    # will be greyed out). Tool bars are generally not greyed out themselves,
    # but the actions within them are.
    enabled = Bool(True)

    # Are the actions, menus and tool bars in this set visible?
    visible = Bool(True)

    # A mapping from human-readable names to globally unique IDs.
    #
    # This mapping is used when interpreting the first item in a location path
    # (i.e., the **path** trait of a **Location** instance).
    #
    # When the path is intepreted, the first component (i.e., the first item
    # before any '/') is checked to see if it is in the mapping, and if so it
    # is replaced with the value in the map.
    #
    # This technique allows paths to start with human readable names, as
    # opposed to IDs (which are required in order to manage the namespace of
    # all action sets).
    #
    # For example, in the Envisage workbench, the menu bar ID is:
    #
    # ``'envisage.workbench.menubar'``
    #
    # Without aliases, you must specify a location like this:
    #
    # ``Location(path='envisage.workbench.menubar/ASubMenu/AGroup')``
    #
    # This is a bit long-winded! Instead, you can define an alias:
    #
    #     ``aliases = { 'MenuBar' : 'envisage.workbench.menubar' }``
    #
    # In that case, you can specify a location like this:
    #
    #     ``Location(path='MenuBar/ASubMenu/AGroup')``
    #
    aliases = Dict(Str, Str)

    #### Trait initializers ###################################################

    def _id_default(self):
        """ Trait initializer. """

        id = '%s.%s' % (type(self).__module__, type(self).__name__)
        logger.warn('action set %s has no Id - using <%s>' % (self, id))

        return id

    def _name_default(self):
        """ Trait initializer. """

        name = camel_case_to_words(type(self).__name__)
        logger.warn('action set %s has no name - using <%s>' % (self, name))

        return name

#### EOF ######################################################################



""" Manages a collection of action sets. """


# Enthought library imports.
from traits.api import HasTraits, List

# Local imports.
from .action_set import ActionSet


class ActionSetManager(HasTraits):
    """ Manages a collection of action sets. """

    #### 'ActionSetManager' interface #########################################

    # The action sets that this manager manages.
    action_sets = List(ActionSet)

    ###########################################################################
    # 'ActionSetManager' interface.
    ###########################################################################

    def get_actions(self, root):
        """ Return all action definitions for a root. """

        return self._get_items(self.action_sets, 'actions', root)

    def get_groups(self, root):
        """ Return all group definitions for a root. """

        return self._get_items(self.action_sets, 'groups', root)

    def get_menus(self, root):
        """ Return all menu definitions for a root. """

        return self._get_items(self.action_sets, 'menus', root)

    def get_tool_bars(self, root):
        """ Return all tool bar definitions for a root. """

        return self._get_items(self.action_sets, 'tool_bars', root)

    ###########################################################################
    # 'Private' interface.
    ###########################################################################

    def _get_items(self, action_sets, attribute_name, root):
        """ Return all actions, groups or menus for a particular root.

        e.g. To get all of the groups::

            self._get_items(action_sets, 'groups', root)

        """

        items = []
        for action_set in action_sets:
            for item in getattr(action_set, attribute_name):
                if self._get_root(item.path, action_set.aliases) == root:
                    items.append(item)

                    # fixme: Hacky, but the model needs to maintain the
                    # action set that contributed the item.
                    item._action_set_ = action_set

                    # fixme: Even hackier if this is a menu then we need to
                    # tag the action set onto all of the groups.
                    if attribute_name in ['menus', 'toolbars']:
                        for group in item.groups:
                            group._action_set_ = action_set

        return items

    def _get_root(self, path, aliases):
        """ Return the effective root for a path.

        If the first component of the path matches an alias, then we return
        the value of the alias.

        e.g. If the aliases are::

            {'MenuBar' : 'envisage.ui.workbench.menubar'}

        and the path is::

           'MenuBar/File/New'

        Then the effective root is::

            'envisage.ui.workbench.menubar'

        If the first component of the path does *not* match an alias, then it
        is returned as is.

        e.g. If the aliases are::

            {'ToolBar' : 'envisage.ui.workbench.toolbar'}

        and the path is::

           'MenuBar/File/New'

        Then the effective root is::

            'MenuBar'

        """

        components = path.split('/')
        if components[0] in aliases:
            root = aliases[components[0]]

        else:
            root = components[0]

        return root

#### EOF ######################################################################



""" Integrates Editra with the an envisage applciation with the remote editor
    plugin.
"""
__author__ = "Enthought"
__version__ = "0.1"

# Standard library imports
import os

# System library imports
import wx

# ETS imports
from traits.api import Any, Bool
from envisage.plugins.remote_editor.editor_plugins.editor_plugin \
        import EditorPlugin

# Editra namespace imports
from wx.tools.Editra.src import ed_menu
from profiler import Profile_Get

# Constants
_ = wx.GetTranslation
ID_RUN_SCRIPT = wx.NewId()
ID_RUN_TEXT = wx.NewId()


class EditraEditorPlugin(EditorPlugin):

    # Client interface

    # Dispatch all command events in same thread as the wx event loop
    ui_dispatch = 'wx'

    # EditraEditorPlugin interface

    # A reference to the Editra mainWindow
    main_window = Any

    # Whether the window should be given focus when it recieves a command
    raise_on_command = Bool(True)

    #--------------------------------------------------------------------------
    # 'EditorPlugin' interface
    #--------------------------------------------------------------------------

    def new(self):
        """ Open a new file in the main window.
        """
        notebook = self.main_window.GetNotebook()
        current = notebook.GetCurrentCtrl()
        if current.GetFileName() != "" or str(current.GetText()) != "":
            notebook.NewPage()
            notebook.GoCurrentPage()
        current.FindLexer('py')
        current.BackSpaceUnIndents = True

        if self.raise_on_command:
            self.main_window.Raise()

    def open(self, filename):
        """ Open the given filename in the main window.
        """
        self.main_window.DoOpen(None, filename)

        if self.raise_on_command:
            self.main_window.Raise()


class RemoteEditorPlugin(object):
    """ Editor plugin for communicating with shells programs, acting as a
        remote editor.
    """

    def __init__(self, Editra=None):
        self.Editra = Editra

    def do_PlugIt(self):
        self.mainWindow = mainWindow = wx.GetApp().GetMainWindow()
        self.log = wx.GetApp().GetLog()
        self.log("[remote editor][info] Installing remote editor plugin")
        menuBar= mainWindow.GetMenuBar()

        # Register the EditorPlugin with the Enthought remote_editor server
        self.client = EditraEditorPlugin(main_window=self.mainWindow)
        self.client.register()

        # Set up a handler for keybindings. Ideally, we would do this through an
        # interface provided by Editra, but since this seems to change with
        # every single Editra release, we take the brute force approach.
        mainWindow.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)

        # Insert menu items
        toolsMenu = menuBar.GetMenuByName("tools")

        mnu_run_text = wx.MenuItem(toolsMenu, ID_RUN_TEXT,
                         _('Execute selection\tShift+Enter'),
                         _('Execute selected text in a python shell'),
                         wx.ITEM_NORMAL)
        mnu_run_text.SetBitmap(wx.Bitmap(os.path.join(
                                    os.path.dirname(__file__),
                                    'images', 'python_runsel_16x16.png')))
        toolsMenu.AppendItem(mnu_run_text, use_bmp=False)

        mnu_run_script = wx.MenuItem(toolsMenu, ID_RUN_SCRIPT,
                         _('Execute script\tCtrl+Enter'),
                         _('Execute file in a python shell'))
        mnu_run_script.SetBitmap(wx.Bitmap(os.path.join(
                                    os.path.dirname(__file__),
                                    'images', 'python_run_16x16.png')))
        toolsMenu.AppendItem(mnu_run_script, use_bmp=False)

        # Bind the events.
        self.mainWindow._handlers['menu'].extend(
            [(ID_RUN_SCRIPT, self.OnRunScript),
                 (ID_RUN_TEXT, self.OnRunText) ])

        # The enable/disable callback for the toolbar button (we need to insert
        # this callback at the front of the stack).
        self.mainWindow._handlers['ui'].insert(0,
            (ID_RUN_TEXT, self.EnableSelection), )

        # Insert toolbar items
        toolBar = mainWindow.GetToolBar()

        self.run_sel_tb = toolBar.AddLabelTool(
            ID_RUN_TEXT, 'Execute selection',
            wx.Bitmap(os.path.join(os.path.dirname(__file__),
                                   'images', 'python_runsel_24x24.png')),
            shortHelp='Execute selection',
            longHelp='Execute selection in shell')

        self.run_file_tb = toolBar.AddLabelTool(
            ID_RUN_SCRIPT, 'Execute',
            wx.Bitmap(os.path.join(os.path.dirname(__file__),
                                   'images', 'python_run_24x24.png')),
            shortHelp='Execute script',
            longHelp='Execute whole file in shell')

        # Just calling AddLabelTool is not displaying the new tools in the
        # toolbar (for Win XP and OS-X at least). Need to call Realize.
        toolBar.Realize()

    def OnRunScript(self, event):
        """ Run the script, prompting for a save if necessary.
        """
        notebook = self.mainWindow.GetNotebook()
        currentTab = notebook.GetCurrentCtrl()
        filename = currentTab.GetFileName()
        if filename == "":
            if self.mainWindow.OnSaveAs(None, page=currentTab):
                filename = currentTab.GetFileName()
            else:
                return
        else:
            if currentTab.GetModify():
                result = notebook.frame.ModifySave()
                if result == wx.ID_CANCEL:
                    return
        self.client.send_command('run_file', filename)

    def OnRunText(self, event):
        """ Run the selected text.
        """
        currentTab = self.mainWindow.GetNotebook().GetCurrentCtrl()
        text = str(currentTab.GetSelectedText()) # Convert from unicode
        if text == "":
            msg = _("Cannot execute. No text is selected.")
            self.mainWindow.PushStatusText(msg)
        else:
            self.client.send_command('run_text', text)

    def OnKeyDown(self, event):
        """ Handles key presses when the version of Editra is less than
            0.3 (no KeyBindings support).
        """
        if event.GetKeyCode() == wx.WXK_RETURN:
            if event.ShiftDown():
                self.OnRunText(event)
            elif event.ControlDown():
                self.OnRunScript(event)
        else:
            event.Skip()

    def EnableSelection(self, evt):
        if not self.mainWindow.IsActive():
            return

        e_id = evt.GetId()
        evt.SetMode(wx.UPDATE_UI_PROCESS_SPECIFIED)
        # Slow the update interval to reduce overhead
        evt.SetUpdateInterval(200)
        ctrl = self.mainWindow.nb.GetCurrentCtrl()
        if e_id == ID_RUN_TEXT:
            evt.Enable(ctrl.GetSelectionStart() != ctrl.GetSelectionEnd())
        else:
            evt.Skip()






""" Tests for the 'Service' trait type. """


# Enthought library imports.
from envisage.api import Application, Plugin, Service
from traits.api import HasTraits, Instance
from traits.testing.unittest_tools import unittest


class TestApplication(Application):
    """ The type of application used in the tests. """

    id = 'test'


class ServiceTestCase(unittest.TestCase):
    """ Tests for the 'Service' trait type. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_service_trait_type(self):
        """ service trait type"""

        class Foo(HasTraits):
            pass

        class PluginA(Plugin):
            id = 'A'
            foo = Instance(Foo, (), service=True)

        class PluginB(Plugin):
            id = 'B'
            foo = Service(Foo)

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])
        application.start()

        # Make sure the services were registered.
        self.assertEqual(a.foo, b.foo)

        # Stop the application.
        application.stop()

        # Make sure the service was unregistered.
        self.assertEqual(None, b.foo)

        # You can't set service traits!
        self.failUnlessRaises(SystemError, setattr, b, 'foo', 'bogus')

        return

    def test_service_trait_type_with_no_service_registry(self):
        """ service trait type with no service registry """

        class Foo(HasTraits):
            pass

        class Bar(HasTraits):
            foo = Service(Foo)

        # We should get an exception because the object does not have an
        # 'service_registry' trait.
        b = Bar()
        self.failUnlessRaises(ValueError, getattr, b, 'foo')

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" Plugin finder for the current package. """


def get_plugins():
    """ Get the plugins from this package. """

    from .orange_plugin import OrangePlugin

    return [OrangePlugin()]

#### EOF #######################################################################




from envisage.ui.action.api import Action, ActionSet, Group


class TextEditorActionSet(ActionSet):
    """ The default action set for the Text Editor plugin.
    """

    groups = [
        Group(
            id = "TextFileGroup",
            path = "MenuBar/File",
            before = "ExitGroup",
        )
    ]

    actions = [
        Action(
            id = "NewFileAction",
            name = "New Text File",
            class_name='envisage.plugins.text_editor.actions.NewFileAction',
            group='TextFileGroup',
            path="MenuBar/File",
        ),

        Action(
            id = 'OpenFile',
            name = "Open Text File...",
            class_name='envisage.plugins.text_editor.actions.OpenFileAction',
            group='TextFileGroup',
            path="MenuBar/File",
        ),
    ]





""" The exception raised when trying to open a non-existent resource. """


class NoSuchResourceError(Exception):
    """ The exception raised when trying to open a non-existent resource. """

    def __init__(self, message=''):
        """ Constructor. """

        Exception.__init__(self, message)

        return

#### EOF ######################################################################



""" Tests for the base extension registry. """


# Enthought library imports.
from envisage.api import Application, ExtensionPoint
from envisage.api import ExtensionRegistry, UnknownExtensionPoint
from traits.api import List
from traits.testing.unittest_tools import unittest


class ExtensionRegistryTestCase(unittest.TestCase):
    """ Tests for the base extension registry. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # We do all of the testing via the application to make sure it offers
        # the same interface!
        self.registry = Application(extension_registry=ExtensionRegistry())

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_empty_registry(self):
        """ empty registry """

        registry = self.registry

        # Make sure there are no extensions.
        extensions = registry.get_extensions('my.ep')
        self.assertEqual(0, len(extensions))

        # Make sure there are no extension points.
        extension_points = registry.get_extension_points()
        self.assertEqual(0, len(extension_points))

        return

    def test_add_extension_point(self):
        """ add extension point """

        registry = self.registry

        # Add an extension *point*.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Make sure there's NO extensions.
        extensions = registry.get_extensions('my.ep')
        self.assertEqual(0, len(extensions))

        # Make sure there's one and only one extension point.
        extension_points = registry.get_extension_points()
        self.assertEqual(1, len(extension_points))
        self.assertEqual('my.ep', extension_points[0].id)

        return

    def test_get_extension_point(self):
        """ get extension point """

        registry = self.registry

        # Add an extension *point*.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Make sure we can get it.
        extension_point = registry.get_extension_point('my.ep')
        self.assertNotEqual(None, extension_point)
        self.assertEqual('my.ep', extension_point.id)

        return

    def test_remove_empty_extension_point(self):
        """ remove empty_extension point """

        registry = self.registry

        # Add an extension point...
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # ...and remove it!
        registry.remove_extension_point('my.ep')

        # Make sure there are no extension points.
        extension_points = registry.get_extension_points()
        self.assertEqual(0, len(extension_points))

        return

    def test_remove_non_empty_extension_point(self):
        """ remove non-empty extension point """

        registry = self.registry

        # Add an extension point...
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # ... with some extensions...
        registry.set_extensions('my.ep', [42])

        # ...and remove it!
        registry.remove_extension_point('my.ep')

        # Make sure there are no extension points.
        extension_points = registry.get_extension_points()
        self.assertEqual(0, len(extension_points))

        # And that the extensions are gone too.
        self.assertEqual([], registry.get_extensions('my.ep'))

        return

    def test_remove_non_existent_extension_point(self):
        """ remove non existent extension point """

        registry = self.registry

        self.failUnlessRaises(
            UnknownExtensionPoint, registry.remove_extension_point, 'my.ep'
        )

        return

    def test_remove_non_existent_listener(self):
        """ remove non existent listener """

        registry = self.registry

        def listener(registry, extension_point, added, removed, index):
            """ Called when an extension point has changed. """

            self.listener_called = (registry, extension_point, added, removed)

            return

        self.failUnlessRaises(
            ValueError, registry.remove_extension_point_listener, listener
        )

        return

    def test_set_extensions(self):
        """ set extensions """

        registry = self.registry

        # Add an extension *point*.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Set some extensions.
        registry.set_extensions('my.ep', [1, 2, 3])

        # Make sure we can get them.
        self.assertEqual([1, 2, 3], registry.get_extensions('my.ep'))

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_extension_point(self, id, trait_type=List, desc=''):
        """ Create an extension point. """

        return ExtensionPoint(id=id, trait_type=trait_type, desc=desc)


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" A test class used to test adapters. """


# Enthought library imports.
from traits.api import Interface


class IFoo(Interface):
    pass

#### EOF ######################################################################



""" A plugin manager that finds plugins in eggs on the 'plugin_path'. """


import logging, pkg_resources, sys
import traceback

from traits.api import Callable, Directory, List, on_trait_change

from .egg_utils import add_eggs_on_path, get_entry_points_in_egg_order
from .plugin_manager import PluginManager


logger = logging.getLogger(__name__)


class EggBasketPluginManager(PluginManager):
    """ A plugin manager that finds plugins in eggs on the 'plugin_path'.

    To declare a plugin (or plugins) in your egg use an entry point in your
    'setup.py' file, e.g.

    [envisage.plugins]
    acme.foo = acme.foo.foo_plugin:FooPlugin
    acme.foo.fred = acme.foo.fred.fred_plugin:FredPlugin

    The left hand side of the entry point declaration MUST be the same as the
    'id' trait of the plugin (e.g. the 'FooPlugin' would have its 'id' trait
    set to 'acme.foo'). This allows the plugin manager to filter out plugins
    using the 'include' and 'exclude' lists (if specified) *without* having to
    import and instantiate them.

    """

    # Entry point Id.
    ENVISAGE_PLUGINS_ENTRY_POINT = 'envisage.plugins'

    #### 'EggBasketPluginManager' protocol #####################################

    # If a plugin cannot be loaded for any reason, this callable is called
    # with the following arguments: entry_point, exception.
    on_broken_plugin = Callable

    def _on_broken_plugin_default(self):
        def handle_broken_plugin(entry_point, exc):
            raise exc
        return handle_broken_plugin

    # If a distribution cannot be loaded for any reason
    # (mainly VersionConflict), this callable is called with the following
    # arguments: distribution, exception.
    on_broken_distribution = Callable

    # A list of directories that will be searched to find plugins.
    plugin_path = List(Directory)

    @on_trait_change('plugin_path[]')
    def _plugin_path_changed(self, obj, trait_name, removed, added):
        self._update_sys_dot_path(removed, added)
        self.reset_traits(['_plugins'])

    # Protected 'PluginManager' protocol ######################################

    def __plugins_default(self):
        """ Trait initializer. """

        plugins = self._harvest_plugins_in_eggs(self.application)

        logger.debug('egg basket plugin manager found plugins <%s>', plugins)

        return plugins

    #### Private protocol #####################################################

    def _create_plugin_from_entry_point(self, ep, application):
        """ Create a plugin from an entry point. """

        klass  = ep.load()
        plugin = klass(application=application)

        # Warn if the entry point is an old-style one where the LHS didn't have
        # to be the same as the plugin Id.
        if ep.name != plugin.id:
            logger.warn(
                'entry point name <%s> should be the same as the '
                'plugin id <%s>' % (ep.name, plugin.id)
            )

        return plugin

    def _get_plugin_entry_points(self, working_set):
        """ Return all plugin entry points in the working set. """

        entry_points = get_entry_points_in_egg_order(
            working_set, self.ENVISAGE_PLUGINS_ENTRY_POINT
        )

        return entry_points

    def _harvest_plugins_in_eggs(self, application):
        """ Harvest plugins found in eggs on the plugin path. """

        # We first add the eggs to a local working set so that when we get
        # the plugin entry points we don't pick up any from other eggs
        # installed on sys.path.
        plugin_working_set = pkg_resources.WorkingSet(self.plugin_path)
        add_eggs_on_path(plugin_working_set, self.plugin_path,
                         self._handle_broken_distributions)

        # We also add the eggs to the global working set as otherwise the
        # plugin classes can't be imported!
        add_eggs_on_path(pkg_resources.working_set, self.plugin_path,
                         self._handle_broken_distributions)

        plugins = []
        for entry_point in self._get_plugin_entry_points(plugin_working_set):
            if self._include_plugin(entry_point.name):
                try:
                    plugin = self._create_plugin_from_entry_point(entry_point,
                                                                  application)
                    plugins.append(plugin)
                except Exception as exc:
                    exc_tb = traceback.format_exc()
                    msg = 'Error loading plugin: %s (from %s)\n%s'\
                        %(entry_point.name, entry_point.dist.location, exc_tb)
                    logger.error(msg)
                    self.on_broken_plugin(entry_point, exc)

        return plugins

    def _handle_broken_distributions(self, errors):
        logger.error('Error loading distributions: %s', errors)
        if self.on_broken_distribution is None:
            raise SystemError('Cannot find eggs %s' % errors)
        else:
            for dist, exc in errors.items():
                self.on_broken_distribution(dist, exc)

    def _update_sys_dot_path(self, removed, added):
        """ Add/remove the given entries from sys.path. """

        for dirname in removed:
            if dirname in sys.path:
                sys.path.remove(dirname)

        for dirname in added:
            if dirname not in sys.path:
                sys.path.append(dirname)

#### EOF ######################################################################



""" The tree editor used in the application browser. """


# Enthought library imports.
from envisage.api import IApplication, IExtensionPoint, IPlugin
from traits.api import Any, HasTraits, List, Str, Undefined
from traitsui.api import TreeEditor, TreeNode

# fixme: non-api imports.
from traitsui.value_tree import value_tree_nodes


class Container(HasTraits):
    """ A container. """

    # The object that contains the container ;^)
    parent = Any

    # The contents of the container.
    contents = List

    # The label.
    label = Str


class IApplicationTreeNode(TreeNode):
    """ A tree node for an Envisage application. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return True

    def get_children(self, obj):
        """ Get the object's children. """

        return [plugin for plugin in obj]

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return IApplication(obj, Undefined) is obj


class IPluginTreeNode(TreeNode):
    """ A tree node for a Envisage plugins. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return False#True

    def get_children(self, obj):
        """ Get the object's children. """

        extension_points = Container(
            label    = 'Extension Points',
            parent   = obj,
            contents = obj.get_extension_points()
        )

        extensions   = Container(
            label    = 'Extensions',
            parent   = obj,
            contents = self._get_extensions(obj)
        )

        return [extension_points, extensions]

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return IPlugin(obj, Undefined) is obj



    def _get_extensions(self, plugin):

        from traitsui.value_tree import ListNode, StringNode

        class MyListNode(ListNode):
            label = Str
            def format_value(self, value):
                return self.label

        extensions = []
        for trait_name, trait in plugin.traits().items():
            if trait.extension_point is not None:
                node = MyListNode(label=trait.extension_point, value=plugin.get_extensions(trait.extension_point))

                extensions.append(node)

        return extensions

class IExtensionPointTreeNode(TreeNode):
    """ A tree node for an extension point. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return False

    def get_children(self, obj):
        """ Get the object's children. """

        return []

    def get_label(self, obj):
        """ Get the object's label. """

        return obj.id

    def when_label_changed(self, obj, callback, remove):
        return

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return IExtensionPoint(obj, Undefined) is obj


class ContainerTreeNode(TreeNode):
    """ A tree node for a Envisage plugins. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return True

    def get_children(self, obj):
        """ Get the object's children. """

        return obj.contents

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return isinstance(obj, Container)



application_browser_tree_nodes  = [
    IApplicationTreeNode(
        auto_open = True,
        label     = 'id',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),

    IExtensionPointTreeNode(
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),

    IPluginTreeNode(
        label     = 'name',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),

    ContainerTreeNode(
        label     = 'label',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),
] + value_tree_nodes

#### EOF ######################################################################



""" A view showing a summary of the running application. """


# Standard library imports.
import inspect

# Enthought library imports.
from envisage.api import IApplication, IExtensionPoint
from envisage.api import IServiceRegistry
from envisage.developer.code_browser.api import CodeBrowser
from apptools.io.api import File
from traits.api import Any, HasTraits, Instance
from traitsui.api import Item, TreeEditor, View

# fixme: non-api import.
from envisage.plugins.text_editor.editor.text_editor import TextEditor

# Local imports.
from .service_registry_browser_tree import \
     service_registry_browser_tree_nodes


service_registry_browser_view = View(
    Item(
        name       = 'service_registry_model',
        show_label = False,
        editor     = TreeEditor(
            nodes       = service_registry_browser_tree_nodes,
            editable    = False,
            orientation = 'vertical',
            hide_root   = True,
            show_icons  = True,
            on_dclick   = 'object.dclick'
        )
    ),

    resizable = True,
    style     = 'custom',
    title     = 'Service Registry',

    width     = .1,
    height    = .1
)


class ServiceRegistryBrowser(HasTraits):
    """ An extension registry browser.

    Actually, this class exists just because to use a trait editor we have
    to have a trait to edit!

    """

    #### 'ServiceRegistryBrowser' interface #################################

    # The application that whose extension registry we are browsing.
    application = Instance(IApplication)

    # The code browser that we use to parse plugin source code.
    code_browser = Instance(CodeBrowser)

    # The extension registry that we are browsing.
    service_registry = Instance(IServiceRegistry)

    # The extension registry that we are browsing.
    service_registry_model = Any#Instance(IServiceRegistry)

    # The workbench service.
    workbench = Instance('envisage.ui.workbench.api.Workbench')

    # The default traits UI view.
    traits_view = service_registry_browser_view

    ###########################################################################
    # 'ServiceRegistryBrowser' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _service_registry_default(self):
        """ Trait initializer. """

        return self.application.service_registry

    def _service_registry_model_default(self):
        """ Trait initializer. """

        from .service_registry_browser_tree import ServiceRegistryModel

        return ServiceRegistryModel(service_registry=self.service_registry)

    def _workbench_default(self):
        """ Trait initializer. """

        workbench = self.application.get_service(
            'envisage.ui.workbench.workbench.Workbench'
        )

        return workbench

    #### Methods ##############################################################

    def dclick(self, obj):
        """ Called when an object in the tree is double-clicked. """

        if hasattr(obj, '_service_id_'):
            protocol = obj._protocol_
            id       = obj._service_id_
            service  = obj.value

            for plugin in self.application:
                if id in plugin._service_ids:
                    self.dclick_service(plugin, protocol, service)
                    break

            else:
                self.workbench.information(
                    'Service not created by a plugin (%s)' % repr(service)
                )

        return

    def dclick_service(self, plugin, protocol, obj):
        """ Called when an extension is double-clicked. """

        # Parse the plugin source code.
        module = self._parse_plugin(plugin)

        # Get the plugin klass.
        klass = self._get_plugin_klass(module, plugin)

        # Edit the plugin.
        editor = self.workbench.edit(
            self._get_file_object(plugin), kind=TextEditor
        )

        # Was the service offered declaratively?
        trait_name = self._get_service_trait(plugin, protocol, obj)
        if trait_name is not None:

            # Does the trait have a default initializer?
            initializer = klass.methods.get('_%s_default' % trait_name)
            if initializer is not None:
                lineno    = initializer.lineno

            else:
                attribute = klass.attributes.get(trait_name)
                lineno    = attribute.lineno

        else:
            lineno = klass.lineno

        editor.select_line(lineno)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_service_trait(self, plugin, protocol, obj):
        """ Return the servicetrait with the specifed Id.

        Return None if the service was not declared via a trait.

        """

        service_traits = plugin.traits(service=True)

        protocol = self.application.import_symbol(protocol)

        for trait_name, trait in service_traits.items():
            if protocol == self._get_service_protocol(trait):
                break

        else:
            trait_name = None

        return trait_name

    def _get_service_protocol(self, trait):
        """ Determine the protocol to register a service trait with. """

        # If a specific protocol was specified then use it.
        if trait.service_protocol is not None:
            protocol = trait.service_protocol

        # Otherwise, use the type of the objects that can be assigned to the
        # trait.
        #
        # fixme: This works for 'Instance' traits, but what about 'AdaptsTo'
        # and 'AdaptedTo' traits?
        else:
            # Note that in traits the protocol can be an actual class or
            # interfacem or the *name* of a class or interface. This allows
            # us to lazy load them!
            protocol = trait.trait_type.klass

        return protocol

    def _get_plugin_klass(self, module, plugin):
        """ Get the klass that defines the plugin. """

        for name, klass in module.klasses.items():
            if name == type(plugin).__name__:
                break

        else:
            klass = None

        return klass

    def _get_file_object(self, obj):
        """ Return a 'File' object for the object's source file. """

        return File(path=inspect.getsourcefile(type(obj)))

    def _parse_plugin(self, plugin):
        """ Parse the plugin source code. """

        filename = self._get_file_object(plugin).path

        return self.code_browser.read_file(filename)

#### EOF ######################################################################



""" An extension registry implementation with multiple providers. """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import List, provides, on_trait_change

# Local imports.
from .extension_registry import ExtensionRegistry
from .i_extension_provider import IExtensionProvider
from .i_provider_extension_registry import IProviderExtensionRegistry


# Logging.
logger = logging.getLogger(__name__)


@provides(IProviderExtensionRegistry)
class ProviderExtensionRegistry(ExtensionRegistry):
    """ An extension registry implementation with multiple providers. """

    #### Protected 'ProviderExtensionRegistry' interface ######################

    # The extension providers that populate the registry.
    _providers = List(IExtensionProvider)

    ###########################################################################
    # 'IExtensionRegistry' interface.
    ###########################################################################

    def set_extensions(self, extension_point_id, extensions):
        """ Set the extensions to an extension point. """

        raise SystemError('extension points cannot be set')

    ###########################################################################
    # 'ProviderExtensionRegistry' interface.
    ###########################################################################

    def add_provider(self, provider):
        """ Add an extension provider. """

        events = self._add_provider(provider)

        for extension_point_id, (refs, added, index) in events.items():
            self._call_listeners(refs, extension_point_id, added, [], index)

        return

    def get_providers(self):
        """ Return all of the providers in the registry. """

        return self._providers[:]

    def remove_provider(self, provider):
        """ Remove an extension provider.

        Raise a 'ValueError' if the provider is not in the registry.

        """

        events = self._remove_provider(provider)

        for extension_point_id, (refs, removed, index) in events.items():
            self._call_listeners(refs, extension_point_id, [], removed, index)

        return

    ###########################################################################
    # Protected 'ExtensionRegistry' interface.
    ###########################################################################

    def _get_extensions(self, extension_point_id):
        """ Return the extensions for the given extension point. """

        # If we don't know about the extension point then it sure ain't got
        # any extensions!
        if not extension_point_id in self._extension_points:
            logger.warn(
                'getting extensions of unknown extension point <%s>' \
                % extension_point_id
            )
            extensions = []

        # Has this extension point already been accessed?
        elif extension_point_id in self._extensions:
            extensions = self._extensions[extension_point_id]

        # If not, then ask each provider for its contributions to the extension
        # point.
        else:
            extensions = self._initialize_extensions(extension_point_id)
            self._extensions[extension_point_id] = extensions

        # We store the extensions as a list of lists, with each inner list
        # containing the contributions from a single provider. Here we just
        # concatenate them into a single list.
        #
        # You could use a list comprehension, here:-
        #
        #     all = [x for y in extensions for x in y]
        #
        # But I'm sure that that makes it any clearer ;^)

        all = []
        for extensions_of_single_provider in extensions:
            all.extend(extensions_of_single_provider)
        return all

    ###########################################################################
    # Protected 'ProviderExtensionRegistry' interface.
    ###########################################################################

    def _add_provider(self, provider):
        """ Add a new provider. """

        # Add the provider's extension points.
        self._add_provider_extension_points(provider)

        # Add the provider's extensions.
        events = self._add_provider_extensions(provider)

        # And finally, tag it into the list of providers.
        self._providers.append(provider)

        return events

    def _add_provider_extensions(self, provider):
        """ Add a provider's extensions to the registry. """

        # Each provider can contribute to multiple extension points, so we
        # build up a dictionary of the 'ExtensionPointChanged' events that we
        # need to fire.
        events = {}

        # Does the provider contribute any extensions to an extension point
        # that has already been accessed?

        for extension_point_id, extensions in self._extensions.items():
            new = provider.get_extensions(extension_point_id)

            # We only need fire an event for this extension point if the
            # provider contributes any extensions.
            if len(new) > 0:
                index = sum(map(len, extensions))
                refs  = self._get_listener_refs(extension_point_id)
                events[extension_point_id] = (refs, new[:], index)

            extensions.append(new)

        return events

    def _add_provider_extension_points(self, provider):
        """ Add a provider's extension points to the registry. """

        for extension_point in provider.get_extension_points():
            self._extension_points[extension_point.id] = extension_point

        return

    def _remove_provider(self, provider):
        """ Remove a provider. """

        # Remove the provider's extensions.
        events = self._remove_provider_extensions(provider)

        # Remove the provider's extension points.
        self._remove_provider_extension_points(provider, events)

        # And finally take it out of the list of providers.
        self._providers.remove(provider)

        return events

    def _remove_provider_extensions(self, provider):
        """ Remove a provider's extensions from the registry. """

        # Each provider can contribute to multiple extension points, so we
        # build up a dictionary of the 'ExtensionPointChanged' events that we
        # need to fire.
        events = {}

        # Find the index of the provider in the provider list. Its
        # contributions are at the same index in the extensions list of lists.
        index = self._providers.index(provider)

        # Does the provider contribute any extensions to an extension point
        # that has already been accessed?
        for extension_point_id, extensions in self._extensions.items():
            old = extensions[index]

            # We only need fire an event for this extension point if the
            # provider contributed any extensions.
            if len(old) > 0:
                offset = sum(map(len, extensions[:index]))
                refs  = self._get_listener_refs(extension_point_id)
                events[extension_point_id] = (refs, old[:], offset)

            del extensions[index]

        return events

    def _remove_provider_extension_points(self, provider, events):
        """ Remove a provider's extension points from the registry. """

        for extension_point in provider.get_extension_points():
            # Remove the extension point.
            del self._extension_points[extension_point.id]

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait change handlers ################################################

    @on_trait_change('_providers:extension_point_changed')
    def _providers_extension_point_changed(self, obj, trait_name, old, event):
        """ Dynamic trait change handler. """

        logger.debug('provider <%s> extension point changed', obj)

        extension_point_id = event.extension_point_id

        # If the extension point has not yet been accessed then we don't fire a
        # changed event.
        #
        # This is because we only access extension points lazily and so we
        # can't tell what has actually changed because we have nothing to
        # compare it to!
        if not extension_point_id in self._extensions:
            return

        # This is a list of lists where each inner list contains the
        # contributions made to the extension point by a single provider.
        #
        # fixme: This causes a problem if the extension point has not yet been
        # accessed! The tricky thing is that if it hasn't been accessed yet
        # how do we know what has changed?!? Maybe we should just return an
        # empty list instead of barfing!
        extensions = self._extensions[extension_point_id]

        # Find the index of the provider in the provider list. Its
        # contributions are at the same index in the extensions list of lists.
        provider_index = self._providers.index(obj)

        # Get the updated list from the provider.
        extensions[provider_index] = obj.get_extensions(extension_point_id)

        # Find where the provider's contributions are in the whole 'list'.
        offset = sum(map(len, extensions[:provider_index]))

        # Translate the event index from one that refers to the list of
        # contributions from the provider, to the list of contributions from
        # all providers.
        index = self._translate_index(event.index, offset)

        # Find out who is listening.
        refs = self._get_listener_refs(extension_point_id)

        # Let any listeners know that the extensions have been added.
        self._call_listeners(
            refs, extension_point_id, event.added, event.removed, index
        )

        return

    #### Methods ##############################################################

    def _initialize_extensions(self, extension_point_id):
        """ Initialize the extensions to an extension point. """

        # We store the extensions as a list of lists, with each inner list
        # containing the contributions from a single provider.
        extensions = []
        for provider in self._providers:
            extensions.append(provider.get_extensions(extension_point_id)[:])

        logger.debug('extensions to <%s> <%s>', extension_point_id, extensions)

        return extensions

    def _translate_index(self, index, offset):
        """ Translate an event index by the given offset. """

        if isinstance(index, slice):
            index = slice(index.start+offset, index.stop+offset, index.step)

        else:
            index = index + offset

        return index

#### EOF ######################################################################



""" A plugin that offers the 'refresh code' functionality. """


# Enthought library imports.
from envisage.api import Plugin
from traits.api import List


class RefreshCodePlugin(Plugin):
    """ A plugin that offers the 'refresh code' functionality. """

    # Extension point Ids.
    ACTION_SETS = 'envisage.ui.workbench.action_sets'

    #### Extension points offered by this plugin ##############################

    # None.

    #### Contributions to extension points made by this plugin ################

    action_sets = List(contributes_to=ACTION_SETS)

    def _action_sets_default(self):
        """ Trait initializer. """

        from envisage.plugins.refresh_code.refresh_code_action_set import (
            RefreshCodeActionSet
        )

        return [RefreshCodeActionSet]

#### EOF ######################################################################



""" A view that allows a developer to browse the current application. """


# Enthought library imports.
from pyface.workbench.api import TraitsUIView
from traits.api import Instance


# The code browser protocol.
CODE_BROWSER = 'envisage.developer.code_browser.api.CodeBrowser'


class ApplicationBrowserView(TraitsUIView):
    """ A view that allows a developer to browse the current application. """

    #### 'IWorkbenchPart' interface ###########################################

    # The part's globally unique identifier.
    id = 'envisage.developer.ui.view.application_browser_view'

    # The part's name (displayed to the user).
    name = 'Plugins'

    #### 'IView' interface ####################################################

    # The category that the view belongs to (this can used to group views when
    # they are displayed to the user).
    category = 'Developer'

    #### 'ApplicationBrowserView' interface ###################################

    # The code browser used to parse Python code.
    code_browser = Instance(CODE_BROWSER)

    ###########################################################################
    # 'TraitsUIView' interface.
    ###########################################################################

    def _obj_default(self):
        """ Trait initializer. """

        # Local imports.
        from .application_browser import ApplicationBrowser

        application_browser = ApplicationBrowser(
            application  = self.window.application,
            code_browser = self.code_browser
        )

        return application_browser

    ###########################################################################
    # 'ApplicationBrowserView' interface.
    ###########################################################################

    def _code_browser_default(self):
        """ Trait initializer. """

        return self.window.application.get_service(CODE_BROWSER)

#### EOF ######################################################################



""" The exception raised when an unknown extension point is referenced. """


class UnknownExtensionPoint(Exception):
    """ The exception raised when an unknown extension point is referenced. """


#### EOF ######################################################################



""" Tests for extension points. """


# Enthought library imports.
from envisage.api import Application, ExtensionPoint
from envisage.api import ExtensionRegistry
from traits.api import HasTraits, Int, List, TraitError
from traits.testing.unittest_tools import unittest


class TestBase(HasTraits):
    """ Base class for all test classes that use the 'ExtensionPoint' type. """

    extension_registry = None


class ExtensionPointTestCase(unittest.TestCase):
    """ Tests for extension  points. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # We do all of the testing via the application to make sure it offers
        # the same interface!
        self.registry = Application(extension_registry=ExtensionRegistry())

        # Set the extension registry used by the test classes.
        TestBase.extension_registry = self.registry

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_invalid_extension_point_type(self):
        """ invalid extension point type """

        # Extension points currently have to be 'List's of something! The
        # default is a list of anything.
        self.failUnlessRaises(TypeError, ExtensionPoint, Int, 'my.ep')

        return

    def test_no_reference_to_extension_registry(self):
        """ no reference to extension registry """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Set the extensions.
        registry.set_extensions('my.ep', 'xxx')

        # Declare a class that consumes the extension.
        class Foo(HasTraits):
            x = ExtensionPoint(List(Int), id='my.ep')

        # We should get an exception because the object does not have an
        # 'extension_registry' trait.
        f = Foo()
        self.failUnlessRaises(ValueError, getattr, f, 'x')

        return

    def test_extension_point_changed(self):
        """ extension point changed """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(id='my.ep')

            def _x_changed(self):
                """ Static trait change handler. """

                self.x_changed_called = True

                return

        f = Foo()

        # Connect the extension points on the object so that it can listen
        # for changes.
        ExtensionPoint.connect_extension_point_traits(f)

        # Set the extensions.
        registry.set_extensions('my.ep', [42, 'a string', True])

        # Make sure that instances of the class pick up the extensions.
        self.assertEqual(3, len(f.x))
        self.assertEqual([42, 'a string', True],  f.x)

        # Make sure the trait change handler was called.
        self.assert_(f.x_changed_called)

        # Reset the change handler flag.
        f.x_changed_called = False

        # Disconnect the extension points on the object.
        ExtensionPoint.disconnect_extension_point_traits(f)

        # Set the extensions.
        registry.set_extensions('my.ep', [98, 99, 100])

        # Make sure the trait change handler was *not* called.
        self.assertEqual(False, f.x_changed_called)

        return

    def test_untyped_extension_point(self):
        """ untyped extension point """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Set the extensions.
        registry.set_extensions('my.ep', [42, 'a string', True])

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(id='my.ep')

        # Make sure that instances of the class pick up the extensions.
        f = Foo()
        self.assertEqual(3, len(f.x))
        self.assertEqual([42, 'a string', True],  f.x)

        g = Foo()
        self.assertEqual(3, len(g.x))
        self.assertEqual([42, 'a string', True],  g.x)

        return

    def test_typed_extension_point(self):
        """ typed extension point """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Set the extensions.
        registry.set_extensions('my.ep', [42, 43, 44])

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(List(Int), id='my.ep')

        # Make sure that instances of the class pick up the extensions.
        f = Foo()
        self.assertEqual(3, len(f.x))
        self.assertEqual([42, 43, 44], f.x)

        g = Foo()
        self.assertEqual(3, len(g.x))
        self.assertEqual([42, 43, 44], g.x)

        return

    def test_invalid_extension_point(self):
        """ invalid extension point """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Set the extensions.
        registry.set_extensions('my.ep', 'xxx')

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(List(Int), id='my.ep')

        # Make sure we get a trait error because the type of the extension
        # doesn't match that of the extension point.
        f = Foo()
        self.failUnlessRaises(TraitError, getattr, f, 'x')

        return

    def test_extension_point_with_no_id(self):
        """ extension point with no Id """

        def factory():
            class Foo(TestBase):
                x = ExtensionPoint(List(Int))

        self.failUnlessRaises(ValueError, factory)

        return

    def test_set_untyped_extension_point(self):
        """ set untyped extension point """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(id='my.ep')

        # Make sure that when we set the trait the extension registry gets
        # updated.
        f = Foo()
        f.x = [42]

        self.assertEqual([42], registry.get_extensions('my.ep'))

        return

    def test_set_typed_extension_point(self):
        """ set typed extension point """

        registry = self.registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Declare a class that consumes the extension.
        class Foo(TestBase):
            x = ExtensionPoint(List(Int), id='my.ep')

        # Make sure that when we set the trait the extension registry gets
        # updated.
        f = Foo()
        f.x = [42]

        self.assertEqual([42], registry.get_extensions('my.ep'))

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_extension_point(self, id, trait_type=List, desc=''):
        """ Create an extension point. """

        return ExtensionPoint(id=id, trait_type=trait_type, desc=desc)


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" A base class for extension registry implementation. """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import Dict, HasTraits, provides

# Local imports.
from .extension_point_changed_event import ExtensionPointChangedEvent
from .i_extension_registry import IExtensionRegistry
from . import safeweakref
from .unknown_extension_point import UnknownExtensionPoint


# Logging.
logger = logging.getLogger(__name__)


@provides(IExtensionRegistry)
class ExtensionRegistry(HasTraits):
    """ A base class for extension registry implementation. """

    ###########################################################################
    # Protected 'ExtensionRegistry' interface.
    ###########################################################################

    # A dictionary of extensions, keyed by extension point.
    _extensions = Dict

    # The extension points that have been added *explicitly*.
    _extension_points = Dict

    # Extension listeners.
    #
    # These are called when extensions are added to or removed from an
    # extension point.
    #
    # e.g. Dict(extension_point, [weakref.ref(callable)])
    #
    # A listener is any Python callable with the following signature:-
    #
    # def listener(extension_registry, extension_point_changed_event):
    #     ...
    _listeners = Dict

    ###########################################################################
    # 'IExtensionRegistry' interface.
    ###########################################################################

    def add_extension_point_listener(self, listener, extension_point_id=None):
        """ Add a listener for extensions being added or removed. """

        listeners = self._listeners.setdefault(extension_point_id, [])
        listeners.append(safeweakref.ref(listener))

        return

    def add_extension_point(self, extension_point):
        """ Add an extension point. """

        self._extension_points[extension_point.id] = extension_point
        logger.debug('extension point <%s> added', extension_point.id)

        return

    def get_extensions(self, extension_point_id):
        """ Return the extensions contributed to an extension point. """

        return self._get_extensions(extension_point_id)[:]

    def get_extension_point(self, extension_point_id):
        """ Return the extension point with the specified Id. """

        return self._extension_points.get(extension_point_id)

    def get_extension_points(self):
        """ Return all extension points. """

        return list(self._extension_points.values())

    def remove_extension_point_listener(self,listener,extension_point_id=None):
        """ Remove a listener for extensions being added or removed. """

        listeners = self._listeners.setdefault(extension_point_id, [])
        listeners.remove(safeweakref.ref(listener))

        return

    def remove_extension_point(self, extension_point_id):
        """ Remove an extension point. """

        self._check_extension_point(extension_point_id)

        # Remove the extension point.
        del self._extension_points[extension_point_id]

        # Remove any extensions to the extension point.
        if extension_point_id in self._extensions:
            old = self._extensions[extension_point_id]
            del self._extensions[extension_point_id]

        else:
            old = []

        refs = self._get_listener_refs(extension_point_id)
        self._call_listeners(refs, extension_point_id, [], old, 0)

        logger.debug('extension point <%s> removed', extension_point_id)

        return

    def set_extensions(self, extension_point_id, extensions):
        """ Set the extensions contributed to an extension point. """

        self._check_extension_point(extension_point_id)

        old = self._get_extensions(extension_point_id)
        self._extensions[extension_point_id] = extensions

        refs = self._get_listener_refs(extension_point_id)
        self._call_listeners(refs, extension_point_id, extensions, old, None)

        return

    ###########################################################################
    # Protected 'ExtensionRegistry' interface.
    ###########################################################################

    def _call_listeners(self, refs, extension_point_id, added, removed, index):
        """ Call listeners that are listening to an extension point. """

        event = ExtensionPointChangedEvent(
            extension_point_id = extension_point_id,
            added              = added,
            removed            = removed,
            index              = index
        )

        for ref in refs:
            listener = ref()
            if listener is not None:
                listener(self, event)

        return

    def _check_extension_point(self, extension_point_id):
        """ Check to see if the extension point exists.

        Raise an 'UnknownExtensionPoint' if it does not.

        """

        if not extension_point_id in self._extension_points:
            raise UnknownExtensionPoint(extension_point_id)

        return

    def _get_extensions(self, extension_point_id):
        """ Return the extensions for the given extension point. """

        return self._extensions.setdefault(extension_point_id, [])

    def _get_listener_refs(self, extension_point_id):
        """ Get weak references to all listeners to an extension point.

        Returns a list containing the weak references to those listeners that
        are listening to this extension point specifically first, followed by
        those that are listening to any extension point.

        """

        refs = []
        refs.extend(self._listeners.get(extension_point_id, []))
        refs.extend(self._listeners.get(None, []))

        return refs

#### EOF ######################################################################
